;/-------------------\
;| GameFixer Plugin  |
;| MaxFPS            |
;|                   |
;| Version 0.01      |
;| Date: 29.06.2012  |
;| PureBasic 4.60    |
;|                   |
;| by Thorium        |
;\-------------------/

EnableExplicit

;Contains constants we need for the DirectDraw hooking.
XIncludeFile "GameFixer DirectDraw Hook.pbi"

;Import all what we need from the GameFixer core.
Import "GameFixer Core.lib"
  UninstallStdProcHook(HookId.i)
  GetOrigStdProcAddr.i(HookId.i)
  InstallDDrawHook.i(*Callback)
  UninstallDDrawHook(DDrawHookId.i)
  InstallDDrawProcHook(*Object, MethodeIndex.i, HookProcAddr.i)
EndImport

;We want to hook the Flip procedure and need a prototype
;to call the original procedure after it is hooked.
Prototype Proto_Flip(*this, *lpDDSurfaceTargetOverride, dwFlags.i)
Global Flip.Proto_Flip

;Holds the hook ID of the DirectDraw hook.
Global DDrawHookId.i

;Holds the hook ID of the Flip procedure hook.
Global FlipHookId.i

;Holds the FPS limit.
Global MaxFps.i

;Needed for delay calculation to limit the FPS.
Global FpsTimerDuration
Global StartTime.i

;We redirect all calls to Flip to this procedure.
;The Flip procedure swaps two surfaces (image buffer) on the graphics card.
;This is typicly used for double buffering: There is a front buffer,
;which is currently displayed and a backbuffer the game is drawing to.
;If the game completed all drawing operations it flips the buffers to
;display the drawn image. And starts over with drawing on the buffer which
;is currently not displayed.
;This is done to prevent flickering, which would occur if the game would
;draw on a buffer while it's displayed.
;So this procedure is called once per frame in most cases. And is a good
;entry point to limit the FPS.
Procedure.i DDraw_Flip_Hook(*this, *lpDDSurfaceTargetOverride, dwFlags.i)
  
  Protected Ret.i
  Protected TimeGab.i
  
  ;Calculates the delay needed to limit the FPS.
  TimeGab = FpsTimerDuration - (timeGetTime_() - StartTime)
  
  ;If a delay is needed for the FPS limiting, we wait the calculated time.
  If TimeGab > 1
    Delay(TimeGab)
  EndIf
  
  ;Calls the original Flip procedure to do the Flip.
  Ret = Flip(*this, *lpDDSurfaceTargetOverride, dwFlags)
  
  ;Gets a new start time for the delay calculation to limit the FPS.
  StartTime = timeGetTime_()
  
  ;Returns the return value of the original Flip procedure.
  ProcedureReturn Ret
  
EndProcedure

;This procedure is called back by a DirectDraw hook.
Procedure DDrawCallback(*Object, Action.i, ObjectType.i)
  
  Select Action
    
    Case #GF_DD_ObjectCreate
      ;A new DirectDraw releated object was been created.
      
      Select ObjectType
          
        ;We only want to hook the Flip procedure, so we need to wait for surface object to be created.
        Case #GF_DD_IDirectDrawSurface, #GF_DD_IDirectDrawSurface2, #GF_DD_IDirectDrawSurface3, #GF_DD_IDirectDrawSurface4, #GF_DD_IDirectDrawSurface7

          If FlipHookId = 0
            
            ;If the Flip hook is not allready installed, we install it.
            FlipHookId  = InstallDDrawProcHook(*Object, #GF_DD_IDDS_Flip, @DDraw_Flip_Hook())

            ;We need to get a new address to call the original Flip procedure,
            ;without getting redirected to DDraw_Flip_Hook.
            ;For easy calling later, we use a prototype here.
            Flip = GetOrigStdProcAddr(FlipHookId)
            
            ;Initializes StartTime for the delay calculation to limit the FPS.
            StartTime = timeGetTime_()
            
          EndIf
          
      EndSelect
    
  EndSelect
  
EndProcedure

;This procedure is called if the plugin DLL is first loaded.
ProcedureDLL AttachProcess(Instance.i)
  
  ;We do nothing here.
  ;It's not needed to include this procedure.
  
EndProcedure

;This procedure is called if the plugin DLL unloads.
ProcedureDLL DetachProcess(Instance.i)
  
  ;If our plugin unloads we need to uninstall the hooks on Flip and DirectDraw.
  UninstallStdProcHook(FlipHookId)
  UnInstallDDrawHook(DDrawHookId)
  
EndProcedure

;This procedure is called by the GameFixer core to configure the plugin.
ProcedureDLL.i Config(Config.s)
  
  ;Installs the hook on DirectDraw.
  ;After that the DDrawCallback procedure is called by the GameFixer core
  ;every time a DirectDraw releated object is created or released.
  DDrawHookId = InstallDDrawHook(@DDrawCallback())
  
  If DDrawHookId = 0
    ;The hooking failed, so we tell the GameFixer core that our plugin failed by returning #False.
    ProcedureReturn #False
  EndIf
  
  ;Sets the FPS limit by the value thats in Config.
  MaxFps = Val(Config)
  If MaxFps < 1
    ;If the FPS limit is 0 or negativ we set it to 60.
    MaxFps = 60
  EndIf
  
  ;Needed for delay calculation to limit the FPS.
  FpsTimerDuration = 1000 / MaxFps
  
  ;We return #True to tell the GameFixer core that all wend fine.
  ProcedureReturn #True
  
EndProcedure

; IDE Options = PureBasic 4.60 (Windows - x86)
; ExecutableFormat = Shared Dll
; CursorPosition = 55
; FirstLine = 27
; Folding = g
; EnableXP
; Executable = B:\GameFixer\Plugins\DDTest.dll
; Warnings = Display
; IncludeVersionInfo
; VersionField0 = 0.1.0.0
; VersionField1 = 0.1.0.0
; VersionField3 = MaxFPS
; VersionField6 = GameFixer Plugin
; VersionField9 = Janos David Ommert
; VersionField15 = VOS_NT_WINDOWS32
; VersionField16 = VFT_DLL